@@ -2,26 +2,28 @@
 
 ;; Copyright (C) 2005, 2006 Free Software Foundation, Inc.
 
-;; Author: Michael Olson (mwolson AT gnu DOT org)
+;; Author: Michael Olson <mwolson@gnu.org>
 
-;; This is free software; you can redistribute it and/or modify it under
-;; the terms of the GNU General Public License as published by the Free
-;; Software Foundation; either version 2, or (at your option) any later
-;; version.
+;; This file is part of EMMS.
+
+;; EMMS is free software; you can redistribute it and/or modify it
+;; under the terms of the GNU General Public License as published by
+;; the Free Software Foundation; either version 2, or (at your option)
+;; any later version.
 ;;
-;; This is distributed in the hope that it will be useful, but WITHOUT
-;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
-;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
-;; for more details.
+;; EMMS is distributed in the hope that it will be useful, but WITHOUT
+;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
+;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
+;; License for more details.
 ;;
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs; see the file COPYING.  If not, write to the
+;; along with EMMS; see the file COPYING.  If not, write to the
 ;; Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 ;; Boston, MA 02110-1301, USA.
 
 ;;; Commentary:
 
-;;; Benefits
+;;; Benefits of MusicPD
 
 ;; MusicPD features crossfade, very little skipping, minor CPU usage,
 ;; many clients, many supported output formats, fast manipulation via
@@ -65,6 +67,11 @@
 ;; (setq emms-player-mpd-server-name "localhost")
 ;; (setq emms-player-mpd-server-port "6600")
 
+;; If your MusicPD setup requires a password, you will need to do the
+;; following.
+;;
+;; (setq emms-player-mpd-server-password "mypassword")
+
 ;; To get track info from MusicPD, do the following.
 ;;
 ;; (add-to-list 'emms-info-functions 'emms-info-mpd)
@@ -95,36 +102,7 @@
 
 (require 'emms-player-simple)
 (require 'emms-source-playlist)  ; for emms-source-file-parse-playlist
-
-(defun emms-player-mpd-get-supported-regexp ()
-  "Returns a regexp of file extensions that MusicPD supports,
-or nil if we cannot figure it out."
-  (let ((out (split-string (shell-command-to-string "mpd --version")
-                           "\n"))
-        (found-start nil)
-        (supported nil))
-    ;; Get supported formats
-    (while (car out)
-      (cond ((string= (car out) "Supported formats:")
-             (setq found-start t))
-            ((and found-start
-                  (not (string= (car out) "")))
-             (setq supported (concat supported (car out)))))
-      (setq out (cdr out)))
-    ;; Create regexp
-    (when (and (stringp supported)
-               (not (string= supported "")))
-      (concat "\\`http://\\|\\.\\(m3u\\|pls\\|"
-              (mapconcat 'identity (delq nil (split-string supported))
-                         "\\|")
-              "\\)\\'"))))
-
-(defvar emms-player-mpd-supported-regexp
-  ;; Use a sane default, just in case
-  (or (emms-player-mpd-get-supported-regexp)
-      (concat "\\`http://\\|"
-              "\\.\\(m3u\\|ogg\\|flac\\|mp3\\|wav\\|mod\\|au\\|aiff\\)\\'"))
-  "Formats supported by MusicPD Client.")
+(require 'tq)
 
 (defgroup emms-player-mpd nil
   "EMMS player for MusicPD."
@@ -149,6 +127,41 @@ leave it set to nil."
                  directory)
   :group 'emms-player-mpd)
 
+(defun emms-player-mpd-get-supported-regexp ()
+  "Returns a regexp of file extensions that MusicPD supports,
+or nil if we cannot figure it out."
+  (let ((out (split-string (shell-command-to-string "mpd --version")
+                           "\n"))
+        (found-start nil)
+        (supported nil))
+    ;; Get supported formats
+    (while (car out)
+      (cond ((string= (car out) "Supported formats:")
+             (setq found-start t))
+            ((and found-start
+                  (not (string= (car out) "")))
+             (setq supported (concat supported (car out)))))
+      (setq out (cdr out)))
+    ;; Create regexp
+    (when (and (stringp supported)
+               (not (string= supported "")))
+      (concat "\\`http://\\|\\.\\(m3u\\|pls\\|"
+              (regexp-opt (delq nil (split-string supported)))
+              "\\)\\'"))))
+
+(defcustom emms-player-mpd-supported-regexp
+  ;; Use a sane default, just in case
+  (or (emms-player-mpd-get-supported-regexp)
+      (concat "\\`http://\\|"
+              "\\.\\(m3u\\|ogg\\|flac\\|mp3\\|wav\\|mod\\|au\\|aiff\\)\\'"))
+  "Formats supported by MusicPD Client."
+  :type 'regexp
+  :set (function
+        (lambda (sym value)
+          (set sym value)
+          (emms-player-set emms-player-mpd 'regex value)))
+  :group 'emms-player-mpd)
+
 (defcustom emms-player-mpd-connect-function 'open-network-stream
   "Function used to initiate the connection to MusicPD.
 It should take same arguments as `open-network-stream' does."
@@ -165,6 +178,12 @@ It should take same arguments as `open-network-stream' does."
   :type '(choice number string)
   :group 'emms-player-mpd)
 
+(defcustom emms-player-mpd-server-password nil
+  "The port of the MusicPD server that we should connect to."
+  :type '(choice (const :tag "None" nil)
+                 string)
+  :group 'emms-player-mpd)
+
 (defcustom emms-player-mpd-check-interval 1
   "How often to check to see whether MusicPD has advanced to the
 next song.  This may be an integer or a floating point number.
@@ -205,108 +224,14 @@ If your EMMS playlist contains stored playlists, set this to nil."
                  'seek
                  'emms-player-mpd-seek)
 
-;;; Transaction Queue (with some improvements) based on tq.el
-
-(defun emms-player-mpd-tq-create (process)
-  "Create and return a transaction queue communicating with PROCESS.
-PROCESS should be a subprocess capable of sending and receiving
-streams of bytes.  It may be a local process, or it may be connected
-to a tcp server on another machine."
-  (let ((tq (cons nil (cons process
-                            (generate-new-buffer
-                             (concat " emms-player-mpd-tq-temp-"
-                                     (process-name process)))))))
-    (set-process-filter process
-                        `(lambda (proc string)
-                           (emms-player-mpd-tq-filter ',tq string)))
-    tq))
-
-;; accessors
-(defun emms-player-mpd-tq-queue (tq)
-  (car tq))
-(defun emms-player-mpd-tq-process (tq)
-  (car (cdr tq)))
-(defun emms-player-mpd-tq-buffer (tq)
-  (cdr (cdr tq)))
-(defun emms-player-mpd-tq-queue-head-question (tq)
-  (car (car (emms-player-mpd-tq-queue tq))))
-(defun emms-player-mpd-tq-queue-head-regexp (tq)
-  (car (cdr (car (emms-player-mpd-tq-queue tq)))))
-(defun emms-player-mpd-tq-queue-head-closure (tq)
-  (car (cdr (cdr (car (emms-player-mpd-tq-queue tq))))))
-(defun emms-player-mpd-tq-queue-head-fn (tq)
-  (cdr (cdr (cdr (car (emms-player-mpd-tq-queue tq))))))
-
-(defun emms-player-mpd-tq-queue-empty (tq)
-  (not (emms-player-mpd-tq-queue tq)))
-
-(defun emms-player-mpd-tq-queue-add (tq question re closure fn)
-  (setcar tq (nconc (emms-player-mpd-tq-queue tq)
-                    (cons (cons question (cons re (cons closure fn))) nil)))
-  'ok)
-
-(defun emms-player-mpd-tq-queue-pop (tq)
-  (setcar tq (cdr (car tq)))
-  (let ((question (emms-player-mpd-tq-queue-head-question tq)))
-    (when question
-      (process-send-string (emms-player-mpd-tq-process tq) question)))
-  (null (car tq)))
-
-(defun emms-player-mpd-tq-enqueue (tq question regexp closure fn)
-  "Add a transaction to transaction queue TQ.
-This sends the string QUESTION to the process that TQ communicates with.
-When the corresponding answer comes back, we call FN
-with two arguments: CLOSURE, and the answer to the question.
-REGEXP is a regular expression to match the entire answer;
-that's how we tell where the answer ends."
-  (let ((sendp (not (emms-player-mpd-tq-queue-head-question tq))))
-    (emms-player-mpd-tq-queue-add tq question regexp closure fn)
-    (when sendp
-      (process-send-string (emms-player-mpd-tq-process tq) question))))
-
-(defun emms-player-mpd-tq-close (tq)
-  "Shut down transaction queue TQ, terminating the process."
-  (delete-process (emms-player-mpd-tq-process tq))
-  (kill-buffer (emms-player-mpd-tq-buffer tq)))
-
-(defun emms-player-mpd-tq-filter (tq string)
-  "Append STRING to the TQ's buffer; then process the new data."
-  (with-current-buffer (emms-player-mpd-tq-buffer tq)
-    (goto-char (point-max))
-    (insert string)
-    (emms-player-mpd-tq-process-buffer tq)))
-
-(defun emms-player-mpd-tq-process-buffer (tq)
-  "Check TQ's buffer for the regexp at the head of the queue."
-  (set-buffer (emms-player-mpd-tq-buffer tq))
-  (if (= 0 (buffer-size)) ()
-    (if (emms-player-mpd-tq-queue-empty tq)
-        (let ((buf (generate-new-buffer "*spurious*")))
-          (copy-to-buffer buf (point-min) (point-max))
-          (delete-region (point-min) (point))
-          (pop-to-buffer buf nil)
-          (error "Spurious communication from process %s, see buffer %s"
-                 (process-name (emms-player-mpd-tq-process tq))
-                 (buffer-name buf)))
-      (goto-char (point-min))
-      (if (re-search-forward (emms-player-mpd-tq-queue-head-regexp tq) nil t)
-          (let ((answer (buffer-substring (point-min) (point))))
-            (delete-region (point-min) (point))
-            (unwind-protect
-                (condition-case nil
-                    (funcall (emms-player-mpd-tq-queue-head-fn tq)
-                             (emms-player-mpd-tq-queue-head-closure tq)
-                             answer)
-                  (error nil))
-              (emms-player-mpd-tq-queue-pop tq))
-            (emms-player-mpd-tq-process-buffer tq))))))
-
 ;;; Dealing with the MusicPD network process
 
 (defvar emms-player-mpd-process nil)
 (defvar emms-player-mpd-queue nil)
 
 (defvar emms-player-mpd-playlist-id nil)
+(make-variable-buffer-local 'emms-player-mpd-playlist-id)
+
 (defvar emms-player-mpd-current-song nil)
 (defvar emms-player-mpd-status-timer nil)
 
@@ -315,21 +240,22 @@ that's how we tell where the answer ends."
   "Regexp that matches the valid status strings that MusicPD can
 return at the end of a request.")
 
-(defun emms-player-mpd-sentinel (proc str)
+(defun emms-player-mpd-sentinel (proc event)
   "The process sentinel for MusicPD."
   (let ((status (process-status proc)))
-    (cond ((memq status '(exit signal closed))
+    (cond ((string-match "^deleted" event)
            (when emms-player-mpd-verbose
-             (message "Closed MusicPD process"))
-           (emms-player-mpd-tq-close emms-player-mpd-queue)
-           (setq emms-player-mpd-queue nil)
-           (setq emms-player-mpd-process nil))
+             (message "MusicPD process was deleted")))
+          ((memq status '(exit signal closed))
+           (emms-player-mpd-close-process t)
+           (when emms-player-mpd-verbose
+             (message "Closed MusicPD process")))
           ((memq status '(run open))
            (when emms-player-mpd-verbose
              (message "MusicPD process started successfully")))
           (t
            (when emms-player-mpd-verbose
-             (message "Other MusicPD status change: %s" status))))))
+             (message "Other MusicPD status change: %s, %s" status event))))))
 
 ;; Ignore a useless byte-compile warning
 (eval-when-compile
@@ -348,10 +274,26 @@ return at the end of a request.")
     (set-process-sentinel emms-player-mpd-process
                           'emms-player-mpd-sentinel)
     (setq emms-player-mpd-queue
-          (emms-player-mpd-tq-create emms-player-mpd-process))
+          (tq-create emms-player-mpd-process))
     (if (fboundp 'set-process-query-on-exit-flag)
         (set-process-query-on-exit-flag emms-player-mpd-process nil)
-      (process-kill-without-query emms-player-mpd-process))))
+      (process-kill-without-query emms-player-mpd-process))
+    ;; send password
+    (when (stringp emms-player-mpd-server-password)
+      (tq-enqueue emms-player-mpd-queue
+                  (concat "password " emms-player-mpd-server-password "\n")
+                  emms-player-mpd-status-regexp nil #'ignore t))))
+
+(defun emms-player-mpd-close-process (&optional from-sentinel)
+  "Terminate the current MusicPD client process.
+FROM-SENTINEL indicates whether this was called by the process sentinel,
+in which case certain checks should not be made."
+  (when (or from-sentinel
+            (and (processp emms-player-mpd-process)
+                 (memq (process-status emms-player-mpd-process) '(run open))))
+    (tq-close emms-player-mpd-queue)
+    (setq emms-player-mpd-queue nil)
+    (setq emms-player-mpd-process nil)))
 
 (defun emms-player-mpd-send (question closure fn)
   "Send the given QUESTION to the MusicPD server.
@@ -359,12 +301,30 @@ When a reply comes, call FN with CLOSURE and the result."
   (emms-player-mpd-ensure-process)
   (unless (string= (substring question -1) "\n")
     (setq question (concat question "\n")))
-  (emms-player-mpd-tq-enqueue emms-player-mpd-queue question
-                              emms-player-mpd-status-regexp
-                              closure fn))
+  (tq-enqueue emms-player-mpd-queue question
+              emms-player-mpd-status-regexp
+              closure fn t))
 
 ;;; Helper functions
 
+(defun emms-player-mpd-get-mpd-filename (file)
+  "Turn FILE into something that MusicPD can understand.
+This usually means removing a prefix."
+  (if (or (not emms-player-mpd-music-directory)
+          (not (eq (aref file 0) ?/))
+          (string-match "\\`http://" file))
+      file
+    (file-relative-name file emms-player-mpd-music-directory)))
+
+(defun emms-player-mpd-get-emms-filename (file)
+  "Turn FILE into something that EMMS can understand.
+This usually means adding a prefix."
+  (if (or (not emms-player-mpd-music-directory)
+          (eq (aref file 0) ?/)
+          (string-match "\\`http://" file))
+      file
+    (expand-file-name file emms-player-mpd-music-directory)))
+
 (defun emms-player-mpd-parse-response (response)
   "Convert the given MusicPD response into a list.
 The car of the list is special:
@@ -444,9 +404,13 @@ The list will be in reverse order."
         (tracks nil))
     (when songs
       (dolist (song-info songs)
-        (let (cell)
-          (when (setq cell (assoc "file" song-info))
-            (let ((track (emms-track 'file (cdr cell))))
+        (let ((file (cdr (assoc "file" song-info))))
+          (when file
+            (setq file (emms-player-mpd-get-emms-filename file))
+            (let* ((type (if (string-match "\\`http://" file)
+                            'url
+                           'file))
+                   (track (emms-track type file)))
               (emms-info-mpd track song-info)
               (setq tracks (cons track tracks)))))))
     (funcall (car closure) (cdr closure) tracks)))
@@ -553,29 +517,32 @@ info from MusicPD."
                      (string-to-number (match-string 1 time)))))
      "time" info)))
 
-(defun emms-player-mpd-sync-from-emms-1 (closure id)
-  (let ((buffer (car closure))
-        (fn (cadr closure))
-        (close (cddr closure)))
-    (when (buffer-live-p buffer)
-      (with-current-buffer buffer
-        (setq emms-player-mpd-playlist-id id))
-      (when (functionp fn)
-        (funcall fn close)))))
+(defun emms-player-mpd-sync-from-emms-1 (closure)
+  (emms-player-mpd-get-playlist-id
+   closure
+   (lambda (closure id)
+     (let ((buffer (car closure))
+           (fn (cdr closure)))
+       (funcall fn buffer id)))))
 
-(defun emms-player-mpd-sync-from-emms (&optional closure callback)
+(defun emms-player-mpd-sync-from-emms (callback)
   "Synchronize the MusicPD playlist with the contents of the
 current EMMS playlist.
-If CALLBACK is provided, call it with CLOSURE once we are done."
+
+If CALLBACK is provided, call it with the current EMMS playlist
+buffer and MusicPD playlist ID when we are done, if there were no
+errors."
   (emms-player-mpd-clear)
   (with-current-emms-playlist
-    (save-excursion
-      (mapc #'emms-player-mpd-add
-            (nreverse
-             (emms-playlist-tracks-in-region (point-min) (point-max)))))
-    (emms-player-mpd-get-playlist-id
-     (cons (current-buffer) (cons callback closure))
-     #'emms-player-mpd-sync-from-emms-1)))
+    (let (tracks)
+      (save-excursion
+        (setq tracks (nreverse
+                      (emms-playlist-tracks-in-region
+                       (point-min) (point-max)))))
+      (emms-player-mpd-add-several-tracks
+       tracks
+       (cons (current-buffer) callback)
+       #'emms-player-mpd-sync-from-emms-1))))
 
 (defun emms-player-mpd-sync-from-mpd-2 (closure info)
   (let ((buffer (car closure))
@@ -587,6 +554,7 @@ If CALLBACK is provided, call it with CLOSURE once we are done."
       (let ((emms-playlist-buffer buffer))
         (with-current-emms-playlist
           (setq emms-player-mpd-playlist-id id)
+          (set-buffer-modified-p nil)
           (if song
               (progn
                 (goto-line (1+ (string-to-number song)))
@@ -620,15 +588,13 @@ MusicPD playlist."
         (time (emms-player-mpd-get-playing-time nil #'ignore info)))
     (cond ((string= status "stop")
            (emms-player-mpd-disconnect t)
-           (if (with-current-emms-playlist
-                 (save-excursion
-                   (forward-line 1)
-                   (emms-playlist-track-at (point))))
-               ;; a track remains, so use the conservative stop method
+           (if song
+               ;; a track remains: the user probably stopped MusicPD
+               ;; manually, so we'll stop EMMS completely
                (let ((emms-player-stopped-p t))
                  (emms-player-stopped))
-             ;; at the last track: we probably ran out of stuff to
-             ;; play, so let EMMS do something further if it wants to
+             ;; no more tracks are left: we probably ran out of things
+             ;; to play, so let EMMS do something further if it wants
              (emms-player-stopped)))
           ((string= status "pause")
            nil)
@@ -645,7 +611,7 @@ MusicPD playlist."
                                        (point))))
              (emms-player-started 'emms-player-mpd)
              (when time
-               (run-hook-with-args 'emms-player-seeked-functions time)))))))
+               (run-hook-with-args 'emms-player-time-set-functions time)))))))
 
 (defun emms-player-mpd-detect-song-change (&optional info)
   "Detect whether a song change has occurred.
@@ -657,15 +623,6 @@ info from MusicPD."
       (emms-player-mpd-detect-song-change-1 nil info)
     (emms-player-mpd-get-status nil #'emms-player-mpd-detect-song-change-1)))
 
-(defun emms-player-mpd-get-filename (file)
-  "Turn FILE into something that MusicPD can understand.
-This usually means removing a prefix."
-  (if (or (null emms-player-mpd-music-directory)
-          (not (eq (aref file 0) ?/))
-          (string-match "\\`http://" file))
-      file
-    (file-relative-name file emms-player-mpd-music-directory)))
-
 (defun emms-player-mpd-quote-file (file)
   "Escape special characters in FILE and surround in double-quotes."
   (concat "\""
@@ -674,8 +631,6 @@ This usually means removing a prefix."
            (emms-replace-regexp-in-string "\\\\" "\\\\\\\\" file))
           "\""))
 
-;;; MusicPD commands
-
 (defun emms-player-mpd-clear ()
   "Clear the playlist."
   (when emms-player-mpd-status-timer
@@ -683,50 +638,112 @@ This usually means removing a prefix."
     (setq emms-player-mpd-status-timer nil))
   (emms-player-mpd-send "clear" nil #'ignore))
 
-(defun emms-player-mpd-add-file (file)
+;;; Adding to the MusicPD playlist
+
+(defun emms-player-mpd-add-file (file closure callback)
   "Add FILE to the current MusicPD playlist.
-If an error occurs, display a relevant message."
-  (setq file (emms-player-mpd-get-filename file))
+If an error occurs, display a relevant message.
+
+Execute CALLBACK with CLOSURE as its first argument when done."
+  (setq file (emms-player-mpd-get-mpd-filename file))
   (emms-player-mpd-send
    (concat "add " (emms-player-mpd-quote-file file))
-   file
-   (lambda (file response)
-     (let ((output (emms-player-mpd-parse-response response)))
-       (when (car output)
-         (message "MusicPD error: %s: %s" file (cdar output)))))))
+   (cons file (cons callback closure))
+   (lambda (closure response)
+     (let ((output (emms-player-mpd-parse-response response))
+           (file (car closure))
+           (callback (cadr closure))
+           (close (cddr closure)))
+       (if (car output)
+           (message "MusicPD error: %s: %s" file (cdar output))
+         (when (functionp callback)
+           (funcall callback close)))))))
 
-(defun emms-player-mpd-add-playlist (playlist)
+(defun emms-player-mpd-add-buffer-contents (closure callback)
+  "Load contents of the current buffer into MusicPD by adding each line.
+This handles both m3u and pls type playlists.
+
+Execute CALLBACK with CLOSURE as its first argument when done."
+  (goto-char (point-min))
+  (let ((format (emms-source-playlist-determine-format)))
+    (when format
+      (emms-player-mpd-add-several-files
+       (emms-source-playlist-files format)
+       closure callback))))
+
+(defun emms-player-mpd-add-playlist (playlist closure callback)
   "Load contents of PLAYLIST into MusicPD by adding each line.
-This handles both m3u and pls type playlists."
+This handles both m3u and pls type playlists.
+
+Execute CALLBACK with CLOSURE as its first argument when done."
   ;; This is useful for playlists of playlists
   (with-temp-buffer
     (insert-file-contents playlist)
-    (goto-char (point-min))
-    (let ((format (emms-source-playlist-determine-format)))
-      (when format
-        (let ((list (emms-source-playlist-files format)))
-          (dolist (file list)
-            (emms-player-mpd-add-file file)))))))
+    (emms-player-mpd-add-buffer-contents closure callback)))
 
-(defun emms-player-mpd-add (track)
-  "Add TRACK to the MusicPD playlist."
+(defun emms-player-mpd-add-streamlist (url closure callback)
+  "Download contents of URL and then add its feeds into MusicPD.
+
+Execute CALLBACK with CLOSURE as its first argument when done."
+  ;; This is useful with emms-streams.el
+  (condition-case nil
+      (progn
+        (require 'url)
+        (with-temp-buffer
+          (url-insert-file-contents url)
+          (emms-player-mpd-add-buffer-contents closure callback)))
+    (error (message (concat "You need to install url.el so that"
+                            " Emms can retrieve this stream")))))
+
+(defun emms-player-mpd-add (track closure callback)
+  "Add TRACK to the MusicPD playlist.
+
+Execute CALLBACK with CLOSURE as its first argument when done."
   (let ((name (emms-track-get track 'name))
         (type (emms-track-get track 'type)))
     (cond ((eq type 'url)
-           (emms-player-mpd-add-file name))
+           (emms-player-mpd-add-file name closure callback))
+          ((eq type 'streamlist)
+           (emms-player-mpd-add-streamlist name closure callback))
           ((or (eq type 'playlist)
                (string-match "\\.\\(m3u\\|pls\\)\\'" name))
-           (emms-player-mpd-add-playlist name))
+           (emms-player-mpd-add-playlist name closure callback))
           ((eq type 'file)
-           (emms-player-mpd-add-file name)))))
+           (emms-player-mpd-add-file name closure callback)))))
+
+(defun emms-player-mpd-add-several-tracks (tracks closure callback)
+  "Add TRACKS to the MusicPD playlist.
+
+Execute CALLBACK with CLOSURE as its first argument when done."
+  (when (consp tracks)
+    (while (cdr tracks)
+      (emms-player-mpd-add (car tracks) nil #'ignore)
+      (setq tracks (cdr tracks)))
+    ;; only execute callback on last track
+    (emms-player-mpd-add (car tracks) closure callback)))
+
+(defun emms-player-mpd-add-several-files (files closure callback)
+  "Add FILES to the MusicPD playlist.
+
+Execute CALLBACK with CLOSURE as its first argument when done."
+  (when (consp files)
+    (while (cdr files)
+      (emms-player-mpd-add-file (car files) nil #'ignore)
+      (setq files (cdr files)))
+    ;; only execute callback on last file
+    (emms-player-mpd-add-file (car files) closure callback)))
 
 ;;; EMMS API
 
 (defun emms-player-mpd-playable-p (track)
   "Return non-nil when we can play this track."
-  (and (memq (emms-track-type track) '(file url playlist))
+  (and (memq (emms-track-type track) '(file url playlist streamlist))
        (string-match (emms-player-get emms-player-mpd 'regex)
-                     (emms-track-name track))))
+                     (emms-track-name track))
+       (condition-case nil
+           (progn (emms-player-mpd-ensure-process)
+                  t)
+         (error nil))))
 
 (defun emms-player-mpd-play (&optional id)
   "Play whatever is in the current MusicPD playlist.
@@ -751,14 +768,17 @@ playlist."
      (lambda (closure response)
        (emms-player-started 'emms-player-mpd)))))
 
-(defun emms-player-mpd-start-and-sync-1 (buffer)
+(defun emms-player-mpd-start-and-sync-1 (buffer id)
   (when emms-player-mpd-status-timer
     (emms-cancel-timer emms-player-mpd-status-timer)
     (setq emms-player-mpd-status-timer nil))
-  (let ((emms-playlist-buffer buffer))
-    (with-current-emms-playlist
-      (emms-player-mpd-play (1- (emms-line-number-at-pos
-                                 emms-playlist-selected-marker))))))
+  (when (buffer-live-p buffer)
+    (let ((emms-playlist-buffer buffer))
+      (with-current-emms-playlist
+        (setq emms-player-mpd-playlist-id id)
+        (set-buffer-modified-p nil)
+        (emms-player-mpd-play (1- (emms-line-number-at-pos
+                                   emms-playlist-selected-marker)))))))
 
 (defun emms-player-mpd-start-and-sync ()
   "Ensure that MusicPD's playlist is up-to-date with EMMS's
@@ -768,12 +788,13 @@ This is called if `emms-player-mpd-sync-playlist' is non-nil."
   (emms-player-mpd-get-playlist-id
    nil
    (lambda (closure id)
-     (if (and (stringp emms-player-mpd-playlist-id)
-              (string= emms-player-mpd-playlist-id id))
-         (emms-player-mpd-start-and-sync-1 emms-playlist-buffer)
-       (emms-player-mpd-sync-from-emms
-        emms-playlist-buffer
-        #'emms-player-mpd-start-and-sync-1)))))
+     (let ((buf-id (with-current-emms-playlist emms-player-mpd-playlist-id)))
+       (if (and (not (buffer-modified-p emms-playlist-buffer))
+                (stringp buf-id)
+                (string= buf-id id))
+           (emms-player-mpd-start-and-sync-1 emms-playlist-buffer id)
+         (emms-player-mpd-sync-from-emms
+          #'emms-player-mpd-start-and-sync-1))))))
 
 (defun emms-player-mpd-connect-1 (closure info)
   (setq emms-player-mpd-current-song nil)
@@ -802,12 +823,13 @@ Afterward, the status of MusicPD will be tracked."
 (defun emms-player-mpd-start (track)
   "Starts a process playing TRACK."
   (interactive)
-  (if emms-player-mpd-sync-playlist
+  (if (and emms-player-mpd-sync-playlist
+           (not (memq (emms-track-get track 'type) '(streamlist playlist)))
+           (not (string-match "\\`http://" (emms-track-get track 'name))))
       (emms-player-mpd-start-and-sync)
     (emms-player-mpd-clear)
-    (when (emms-player-mpd-add track)
-      ;; if we have loaded the item successfully, play it
-      (emms-player-mpd-play))))
+    ;; if we have loaded the item successfully, play it
+    (emms-player-mpd-add track nil #'emms-player-mpd-play)))
 
 (defun emms-player-mpd-disconnect (&optional no-stop)
   "Terminate the MusicPD client process and disconnect from MusicPD.
@@ -818,8 +840,8 @@ from other functions."
   (interactive)
   (emms-cancel-timer emms-player-mpd-status-timer)
   (setq emms-player-mpd-status-timer nil)
-  (setq emms-player-mpd-playlist-id nil)
   (setq emms-player-mpd-current-song nil)
+  (emms-player-mpd-close-process)
   (unless no-stop
     (let ((emms-player-stopped-p t))
       (emms-player-stopped))))
@@ -827,10 +849,10 @@ from other functions."
 (defun emms-player-mpd-stop ()
   "Stop the currently playing song."
   (interactive)
-  (emms-player-mpd-disconnect t)
   (condition-case nil
       (emms-player-mpd-send "stop" nil #'ignore)
     (error nil))
+  (emms-player-mpd-disconnect t)
   (let ((emms-player-stopped-p t))
     (emms-player-stopped)))
 
@@ -839,13 +861,18 @@ from other functions."
   (interactive)
   (emms-player-mpd-send "pause" nil #'ignore))
 
-(defun emms-player-mpd-seek (sec)
-  "Seek backward or forward by SEC seconds, depending on sign of SEC."
+(defun emms-player-mpd-seek (amount)
+  "Seek backward or forward by AMOUNT seconds, depending on sign of AMOUNT."
   (interactive)
-  (emms-player-mpd-send (format "seek %s%d"
-                                (if (> sec 0) "+" "")
-                                sec)
-                        nil #'ignore))
+  (emms-player-mpd-get-status
+   amount
+   (lambda (amount info)
+     (let ((song (emms-player-mpd-get-current-song nil #'ignore info))
+           (secs (emms-player-mpd-get-playing-time nil #'ignore info)))
+       (when (and song secs)
+         (emms-player-mpd-send
+          (concat "seek " song " " (number-to-string (+ secs amount)))
+          nil #'ignore))))))
 
 (defun emms-player-mpd-next ()
   "Move forward by one track in MusicPD's internal playlist."
@@ -857,7 +884,9 @@ from other functions."
   (interactive)
   (emms-player-mpd-send "previous" nil #'ignore))
 
-(defun emms-player-mpd-volume-change (amount)
+;;; Volume
+
+(defun emms-volume-mpd-change (amount)
   "Change volume up or down by AMOUNT, depending on whether it is
 positive or negative."
   (interactive "MVolume change amount (+ increase, - decrease): ")
@@ -869,37 +898,52 @@ positive or negative."
         (concat "setvol \"" (number-to-string new-volume) "\"")
         nil #'ignore)))))
 
-(defun emms-player-mpd-volume-up ()
+(defun emms-volume-mpd-raise ()
   "Increase the volume."
   (interactive)
-  (emms-player-mpd-volume-change 5))
+  (emms-volume-mpd-change 5))
 
-(defun emms-player-mpd-volume-down ()
+(defun emms-volume-mpd-lower ()
   "Decrease the volume."
   (interactive)
-  (emms-player-mpd-volume-change -5))
+  (emms-volume-mpd-change -5))
+
+;;; Now playing
 
 (defun emms-player-mpd-show-1 (closure response)
   (let* ((info (emms-player-mpd-get-alist
                 (emms-player-mpd-parse-response response)))
-         (track (emms-dictionary '*track*))
          (insertp (car closure))
          (callback (cadr closure))
          (buffer (cddr closure))
+         (name (cdr (assoc "name" info))) ; radio feeds sometimes set this
+         (file (cdr (assoc "file" info)))
          (desc nil))
     (when info
-      (emms-track-set track 'type 'file)
-      (emms-track-set track 'name (cdr (assoc "file" info)))
-      (emms-info-mpd track info)
-      (setq desc (emms-track-description track)))
+      (when name
+        (setq desc name))
+      (when file
+        (let ((track (emms-dictionary '*track*))
+              track-desc)
+          (if (string-match "\\`http://" file)
+              (emms-track-set track 'type 'url)
+            (emms-track-set track 'type 'file))
+          (emms-track-set track 'name file)
+          (emms-info-mpd track info)
+          (setq track-desc (emms-track-description track))
+          (when (and (stringp track-desc) (not (string= track-desc "")))
+            (setq desc (if desc
+                           (concat desc ": " track-desc)
+                         track-desc))))))
     (if (not desc)
         (message "Nothing playing right now")
       (setq desc (format emms-show-format desc))
       (cond ((functionp callback)
              (funcall callback buffer desc))
             (insertp
-             (with-current-buffer buffer
-               (insert desc)))
+             (when (buffer-live-p buffer)
+               (with-current-buffer buffer
+                 (insert desc))))
             (t
              (message "%s" desc))))))
 
@@ -949,20 +993,93 @@ If INFO is specified, use that instead of acquiring the necessary
 info from MusicPD."
   (if info
       (emms-info-mpd-process track info)
-    (let (file)
-      (when (and (eq 'file (emms-track-type track))
-                 emms-player-mpd-music-directory
-                 (setq file (emms-player-mpd-get-filename
-                             (emms-track-name track)))
-                 (string-match emms-player-mpd-supported-regexp file)
-                 (not (string-match "\\`http://" file)))
-        (condition-case nil
-            (emms-player-mpd-send
-             (concat "find filename "
-                     (emms-player-mpd-quote-file file))
-             track
-             #'emms-info-mpd-1)
-          (error nil))))))
+    (when (and (eq 'file (emms-track-type track))
+               (not (string-match "\\`http://" file)))
+      (let ((file (emms-player-mpd-get-mpd-filename (emms-track-name track))))
+        (when (or emms-player-mpd-music-directory
+                  (and file
+                       (string-match emms-player-mpd-supported-regexp file)))
+          (condition-case nil
+              (emms-player-mpd-send
+               (concat "find filename "
+                       (emms-player-mpd-quote-file file))
+               track
+               #'emms-info-mpd-1)
+            (error nil)))))))
+
+;;; Caching
+
+(defun emms-cache-set-from-mpd-track (track-info)
+  "Dump TRACK-INFO into the EMMS cache.
+The track should be an alist as per `emms-player-mpd-get-alist'."
+  (when emms-cache-set-function
+    (let ((track (emms-dictionary '*track*))
+          (name (cdr (assoc "file" track-info))))
+      (when name
+        (setq name (emms-player-mpd-get-emms-filename name))
+        (emms-track-set track 'type 'file)
+        (emms-track-set track 'name name)
+        (emms-info-mpd-process track track-info)
+        (funcall emms-cache-set-function 'file name track)))))
+
+(defun emms-cache-set-from-mpd-directory (dir)
+  "Dump all MusicPD data from DIR into the EMMS cache.
+This is useful to do when you have recently acquired new music."
+  (interactive
+   (list (if emms-player-mpd-music-directory
+             (emms-read-directory-name "Directory: "
+                                       emms-player-mpd-music-directory)
+           (read-string "Directory: "))))
+  (unless (string= dir "")
+    (setq dir (emms-player-mpd-get-mpd-filename dir)))
+  (if emms-cache-set-function
+      (progn
+        (message "Dumping MusicPD data to cache...")
+        (emms-player-mpd-send
+         (concat "listallinfo " dir)
+         nil
+         (lambda (closure response)
+           (message "Dumping MusicPD data to cache...processing")
+           (let ((info (emms-player-mpd-get-alists
+                        (emms-player-mpd-parse-response response))))
+             (dolist (track-info info)
+               (emms-cache-set-from-mpd-track track-info))
+             (message "Dumping MusicPD data to cache...done")))))
+    (error "Caching is not enabled")))
+
+(defun emms-cache-set-from-mpd-all ()
+  "Dump all MusicPD data into the EMMS cache.
+This is useful to do once, just before using emms-browser.el, in
+order to prime the cache."
+  (interactive)
+  (emms-cache-set-from-mpd-directory ""))
+
+;;; Updating tracks
+
+(defun emms-player-mpd-update-directory (dir)
+  "Cause the tracks in DIR to be updated in the MusicPD database."
+  (interactive
+   (list (if emms-player-mpd-music-directory
+             (emms-read-directory-name "Directory: "
+                                       emms-player-mpd-music-directory)
+           (read-string "Directory: "))))
+  (unless (string= dir "")
+    (setq dir (emms-player-mpd-get-mpd-filename dir)))
+  (emms-player-mpd-send
+   (concat "update " dir) nil
+   (lambda (closure response)
+     (let ((id (cdr (assoc "updating_db"
+                           (emms-player-mpd-get-alist
+                            (emms-player-mpd-parse-response response))))))
+       (if id
+           (message "Updating DB with ID %s" id)
+         (message "Could not update the DB"))))))
+
+(defun emms-player-mpd-update-all ()
+  "Cause all tracks in the MusicPD music directory to be updated in
+the MusicPD database."
+  (interactive)
+  (emms-player-mpd-update-directory ""))
 
 (provide 'emms-player-mpd)
 
