@@ -1,21 +1,24 @@
-;;; emms-stream-info.el --- Show what is currently playing on a
-;;; streaming audio station.
+;;; emms-stream-info.el --- Info from streaming audio
 
-;; Copyright (C) 2004  Yoni Rabkin Katzenell <yoni-r@actcom.com>
-;; This program is free software; you can redistribute it and/or
-;; modify it under the terms of the GNU General Public License
-;; as published by the Free Software Foundation; either version 2
-;; of the License, or (at your option) any later version.
+;; Copyright (C) 2004, 2005, 2006 Free Software Foundation, Inc.
 
-;; This program is distributed in the hope that it will be useful,
+;; Author: Yoni Rabkin <yonirabkin@member.fsf.org>
+
+;; This file is part of EMMS.
+
+;; EMMS is free software; you can redistribute it and/or
+;; modify it under the terms of the GNU General Public License as
+;; published by the Free Software Foundation; either version 2 of the
+;; License, or (at your option) any later version.
+
+;; EMMS is distributed in the hope that it will be useful,
 ;; but WITHOUT ANY WARRANTY; without even the implied warranty of
 ;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with this program; if not, write to the Free Software
-;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-;; 02111-1307, USA.
+;; along with EMMS; if not, write to the Free Software Foundation,
+;; Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
 
 ;;; Commentary:
 ;;
@@ -63,10 +66,12 @@
 ;;    channels explicitly request 3rd party applications not to
 ;;    connect to their service. This is their prerogative. Respect it.
 
-;; $Id: emms-stream-info.el,v 1.7 2005/06/17 10:19:08 yonirabkin Exp $
+;; $Id: emms-stream-info.el,v 1.8 2005/07/09 11:56:00 forcer Exp $
 
 ;;; Code:
 
+(require 'emms)
+
 ;; A higher value for 'emms-stream-info-max' this gives us a
 ;; correspondingly higher chance of grabbing the title information
 ;; from a stream but incurs a price in the additional time it takes to
@@ -87,7 +92,7 @@
   "Output real-time information about the connection.")
 
 (defconst emms-stream-info-version
-  "$Revision: 1.7 $"
+  "$Revision: 1.8 $"
   "Software version.")
 
 (defconst emms-stream-info-char-alter-regexp "[-,'=:%+&0-9A-Za-z\.()/ ]"
@@ -114,9 +119,9 @@
   (concat emms-stream-info-char-alter-regexp "+")
   "Regular expression for metainformation headers.")
 
-(defconst emms-stream-info-playlist-regexp
+(defconst emms-stream-info-streamlist-regexp
   "\\(^http://.*\\)\\|^File.=\\(http://.*\\)"
-  "Regular expression for playlist URLs.")
+  "Regular expression for streamlist URLs.")
 
 ;; When t output debugging info
 (defconst emms-stream-info-debugging nil
@@ -158,10 +163,10 @@ and songtitle information. This string specifies an alternate
 format for those stations.")
 
 (defconst emms-stream-info-pls-regexp ".*\.pls"
-  "Regular expression for a .pls playlist file.")
+  "Regular expression for a .pls streamlist file.")
 
 (defconst emms-stream-info-m3u-regexp ".*\.m3u"
-  "Regular expression for a .m3u playlist file.")
+  "Regular expression for a .m3u streamlist file.")
 
 (defvar emms-stream-info-url nil
   "Server URL.")
@@ -172,8 +177,8 @@ format for those stations.")
 (defvar emms-stream-info-found nil
   "Results of our search.")
 
-(defvar emms-stream-info-playlist-found nil
-  "Results of our playlist search.")
+(defvar emms-stream-info-streamlist-found nil
+  "Results of our streamlist search.")
 
 (defvar emms-stream-info-procname "emms-stream-info-process"
   "Name of network connection process.")
@@ -198,8 +203,8 @@ functions use 'emms-stream-info-return-hook'.")
 (defvar emms-stream-info-title-flag nil
   "Non-nil means title information has been captured.")
 
-(defvar emms-stream-info-playlist-flag nil
-  "Non-nil means playlist information has been captured.")
+(defvar emms-stream-info-streamlist-flag nil
+  "Non-nil means streamlist information has been captured.")
 
 (defvar emms-stream-info-request-string nil
   "String sent to streaming audio server.")
@@ -240,7 +245,9 @@ functions use 'emms-stream-info-return-hook'.")
 ;; This is our tiny state machine for keeping track across multiple
 ;; connections.
 (defvar emms-stream-info-state-bv
-  (make-bool-vector 3 nil)
+  (if (fboundp 'make-bool-vector)
+      (make-bool-vector 3 nil)
+    (make-vector 3 nil))
   "State of sequential connections.
 true at index 0 means output formatted message.
 true at index 1 means insert formatted message.
@@ -270,7 +277,7 @@ trye at index 2 means continue to next connection.")
 (defun emms-stream-info-continue-p ()
   (aref emms-stream-info-state-bv 2))
 
-(defun emms-stream-info-playlist-type (str)
+(defun emms-stream-info-streamlist-type (str)
   (if (stringp str)
       (cond ((string-match emms-stream-info-pls-regexp str)
 	     'pls)
@@ -284,7 +291,7 @@ trye at index 2 means continue to next connection.")
     (setq key-list (mapcar 'car format-alist))
     (mapc (lambda (e)
 	    (setq str 
-		  (replace-regexp-in-string 
+		  (emms-replace-regexp-in-string 
 		   e 
 		   (cdr (assoc e format-alist))
 		   str)))
@@ -320,7 +327,7 @@ not output a message and only return a string."
 	   format-alist))
     
     ;; Escape rougue percent signs hiding in our string.
-    (setq str (replace-regexp-in-string "%" "%%" str))
+    (setq str (emms-replace-regexp-in-string "%" "%%" str))
 
     ;; Either output a message or return a string. But only if it is
     ;; an identifiable station/channel
@@ -335,12 +342,12 @@ not output a message and only return a string."
 
 (defun emms-stream-info-continue ()
   (emms-stream-info-unset-continue)
-  (if emms-stream-info-playlist-found
-      (emms-stream-info-get emms-stream-info-playlist-found
+  (if emms-stream-info-streamlist-found
+      (emms-stream-info-get emms-stream-info-streamlist-found
 			    (emms-stream-info-message-p)
 			    (emms-stream-info-insert-p)
 			    nil)
-    (error "No playlist found at URL")))
+    (error "No streamlist found at URL")))
 
 ;; Useful
 (defun list-to-string (l)
@@ -375,13 +382,13 @@ KEYS. KEYS should be a list of strings."
     (re-search-forward (concat "artist=\\(" 
 			       emms-stream-info-icecast-regexp 
 			       "\\)") end t)
-    (setq artist (match-string-no-properties 1))
+    (setq artist (emms-match-string-no-properties 1))
     
     (goto-char begin)
     (re-search-forward (concat "title=\\(" 
 			       emms-stream-info-icecast-regexp 
 			       "\\)") end t)
-    (setq title (match-string-no-properties 1))
+    (setq title (emms-match-string-no-properties 1))
 
     ;; ugh
     (if (or artist title)
@@ -446,7 +453,7 @@ Argument STR Quanta of data."
 		    (progn
 		      (add-to-list 'emms-stream-info-found 
 				   (cons term 
-					 (match-string-no-properties 1)))
+					 (emms-match-string-no-properties 1)))
 		      (setq emms-stream-info-header-flag t))))
 	      emms-stream-info-vocab))
 
@@ -480,16 +487,16 @@ Argument STR Quanta of data."
 
     ;; Too many nested conditions
     (if (emms-stream-info-set-continue)
-	(unless emms-stream-info-playlist-flag
+	(unless emms-stream-info-streamlist-flag
 	  (goto-char (point-min))
 	  (if (re-search-forward  
-	       emms-stream-info-playlist-regexp
+	       emms-stream-info-streamlist-regexp
 	       (point-max) t)
 	      (progn
-		(setq emms-stream-info-playlist-found
-		      (or (match-string-no-properties 1)
-			  (match-string-no-properties 2)))
-		(setq emms-stream-info-playlist-flag t))))))
+		(setq emms-stream-info-streamlist-found
+		      (or (emms-match-string-no-properties 1)
+			  (emms-match-string-no-properties 2)))
+		(setq emms-stream-info-streamlist-flag t))))))
 
   ;; Be chatty at the user
   (if emms-stream-info-verbose
@@ -502,8 +509,8 @@ Argument STR Quanta of data."
   (if (or (> emms-stream-info-downloaded emms-stream-info-max) ; maxed out?
 	  ;; Captured header and title info?
 	  (and emms-stream-info-header-flag emms-stream-info-title-flag)
-	  ;; Captured playlist info?
-	  emms-stream-info-playlist-flag)
+	  ;; Captured streamlist info?
+	  emms-stream-info-streamlist-flag)
       (emms-stream-info-kill-process proc)))
 
 ;; Closing the connection proves to be the most difficult part of the
@@ -590,8 +597,8 @@ server at URLSTRING."
   (setq emms-stream-info-title-flag nil)	 ; forget title flag
   (setq emms-stream-info-header-flag nil)	 ; forget header flag
   (setq emms-stream-info-found nil)		 ; forget output
-  (setq emms-stream-info-playlist-found nil)     ; forget playlist
-  (setq emms-stream-info-playlist-flag nil)      ; forget playlist
+  (setq emms-stream-info-streamlist-found nil)	 ; forget streamlist
+  (setq emms-stream-info-streamlist-flag nil)	 ; forget streamlist
   (setq emms-stream-info-read-inhibit t)         ; do not read output
 
   ;; Reset state machine
@@ -622,9 +629,11 @@ Optional argument CONT boolean."
 
   (if urlstring
       (emms-stream-info-parse-url urlstring)
-    (emms-stream-info-parse-url 
-     ;; possible bug, what if there is no last stream?
-     (emms-stream-url emms-stream-last-stream)))
+    (and (boundp 'emms-stream-last-stream)
+         (fboundp 'emms-stream-url)
+         emms-stream-last-stream
+         (emms-stream-info-parse-url
+          (emms-stream-url emms-stream-last-stream))))
 
   (emms-stream-info-reset-state)
 
@@ -689,16 +698,21 @@ Optional argument CONT boolean."
       (unless (process-sentinel proc)
 	(error "No process sentinel")))))
 
-;; Users. You can never tell what they are going to use as input.
+;; Should be phased out.
+;; (defun emms-stream-info-input-sanity (&optional urlstring)
+;;   (let ((type (emms-track-type (emms-playlist-selected-track))))
+;;     (cond ((null urlstring)
+;; 	   (if (or (equal type 'streamlist)
+;; 		   (equal type 'url))
+;; 	       (emms-track-name (emms-playlist-selected-track))))
+;; 	  ((not (stringp urlstring))
+;; 	   (error "URL must be in string format"))
+;; 	  ((stringp url) urlstring))))
+
 (defun emms-stream-info-input-sanity (&optional urlstring)
-  (let ((type (emms-track-type (emms-playlist-current-track))))
-    (cond ((null urlstring)
-	   (if (or (equal type 'playlist)
-		   (equal type 'url))
-	       (emms-track-name (emms-playlist-current-track))))
-	  ((not (stringp urlstring))
-	   (error "URL must be in string format"))
-	  ((stringp url) urlstring))))
+  (if (stringp urlstring)
+      urlstring
+    (error "URL must be in string format")))
 
 (defun emms-stream-info-message (&optional urlstring)
   "Get information from streaming audio server at URLSTRING.
@@ -706,9 +720,9 @@ Return a formatted message.
 URLSTRING should be a string."
   (interactive)
   (let ((url (emms-stream-info-input-sanity urlstring)))
-    (cond ((equal (emms-stream-info-playlist-type url) 'pls)
+    (cond ((equal (emms-stream-info-streamlist-type url) 'pls)
 	   (emms-stream-info-get url t nil t))
-	  ((equal (emms-stream-info-playlist-type url) 'm3u)
+	  ((equal (emms-stream-info-streamlist-type url) 'm3u)
 	   (emms-stream-info-get url t nil t))
 	  (t (emms-stream-info-get url t)))))
 
@@ -719,9 +733,9 @@ Insert a formatted message at point.
 URLSTRING should be a string."
   (interactive)
   (let ((url (emms-stream-info-input-sanity urlstring)))
-    (cond ((equal (emms-stream-info-playlist-type url) 'pls)
+    (cond ((equal (emms-stream-info-streamlist-type url) 'pls)
 	   (emms-stream-info-get url nil t t))
-	  ((equal (emms-stream-info-playlist-type url) 'm3u)
+	  ((equal (emms-stream-info-streamlist-type url) 'm3u)
 	   (emms-stream-info-get url nil t t))
 	  (t (emms-stream-info-get url nil t)))))
 
